/*!
 * jGoBoard v1.0
 * http://www.jgoboard.com/
 *
 * This software is licensed under a Creative Commons Attribution-NonCommercial 3.0 Unported License:
 * http://creativecommons.org/licenses/by-nc/3.0/
 * If you want to negotiate on a commercial license, please contact the author.
 *
 * Date: 2011-02-17
 */

/**
 * JGONavigator is a helper class designed to progress a game tree forwards, backwards and traverse variations.
 */
function JGONavigator(gameTree) {
	this.root = gameTree; // save for fast access to beginning
	this.curTree = gameTree; // also current game tree
	this.seqPos = 0; // current position in the sequence
	this.treeStack = []; // we push the parent when we enter a leaf
	this.branches = []; // gametree indices followed
	this.position = 0; // overall position
}

/**
 * Advance to the next node in the game tree.
 *
 * @param {int} variation (Optional) parameter to specify which variation to select, if there are several branches
 * @returns {boolean} True if advance was successful, false if we are at the end of the current variation
 */
JGONavigator.prototype.next = function(variation) {
	if(variation == undefined)
		variation = 0;
		
	if(this.seqPos+1 < this.curTree.sequence.length) { // still nodes in current tree's sequence
		this.seqPos++;
		this.position++;
		return true;
	} else if(this.curTree.leaves.length == 0) { // end of game
		return false;
	} else { // 1 to N subtrees
		var n = this.curTree.leaves.length; // number of variations
		
		this.treeStack.push(this.curTree); // save current tree
		
		this.curTree = this.curTree.leaves[variation]; // advance
		this.branches.push(variation); 
		this.seqPos = 0;
		this.position++;
		
		return true;
	}
};

/**
 * Back up a node in the game tree.
 *
 * @returns {boolean} True if successful, false if we are at the beginning of the game tree
 */
JGONavigator.prototype.previous = function() {
	if(this.seqPos-1 >= 0) { // space to back up in current tree node
		this.seqPos--;
		this.position--;
		return true;
	} else if(this.treeStack.length == 0) { // if already at very beginning
		return false;
	} else { // time to pop up previous tree
		this.curTree = this.treeStack.pop(); // up one branch
		this.branches.pop(); // throw away last branch index - TODO: change this and next() to remember the branch!
		this.seqPos = this.curTree.sequence.length - 1; // go to last item
		this.position--;
		
		return true;
	}
}


/**
 * Go to the beginning of the game tree.
 */
JGONavigator.prototype.reset = function() {
	this.curTree = this.root;
	this.seqPos = 0; 
	this.position = 0;
	this.treeStack = [];
	this.branches = [];
}

/**
 * Get current node. Fails with SGF files with multiple top-level game trees (i.e. no nodes at all in the root tree, only branches).
 *
 * @returns {Object} Node object.
 */
JGONavigator.prototype.get = function() {
	return this.curTree.sequence[this.seqPos];
}

/**
 * Get the overall position in the game tree.
 *
 * @returns {int} Zero-based position in the game tree (nodes from the beginning)
 */
JGONavigator.prototype.getPosition = function() {
	return this.position;
}


/**
 * Get the number of sibling nodes for the current node. Used to browse variations for the current node.
 *
 * @returns {int} Number of sibling leaves (excluding current one), or zero if we are not in the beginning of a branch.
 */
JGONavigator.prototype.getVariations = function() {
	if(this.seqPos != 0 || this.treeStack.length == 0) // not at beginning of a branch, or at the very beginning of game tree
		return 0;
	
	return this.treeStack[this.treeStack.length - 1].leaves.length - 1; // previous tree leaves - 1
}

/**
 * Get the current variation number - returns the latest branch.
 *
 * @returns {int} Number of current variation (zero-based)
 */
JGONavigator.prototype.getVariation = function() {
	if(this.branches.length == 0)
		return 0;
	
	return this.branches[this.branches.length - 1];
}

/**
 * Get the comment related to current node, if any.
 * @param {string} linebreak (Optional) parameter to specify line break identifier - defaults to \n
 * @returns {string} Comment or empty string
 */
JGONavigator.prototype.getComment = function(linebreak) {
	var node = this.get();
	
	if("C" in node) {
		var comment = node["C"][0];
		
		if(linebreak != undefined)
			comment = comment.replace(/\n/g, linebreak);
		
		return comment;
	}
	else
		return "";
}

/**
 * Check if we are at the end of the current variation.
 *
 * @returns {boolean} True if at end, false if not.
 */
JGONavigator.prototype.isAtEnd = function() {
	return (this.seqPos+1 == this.curTree.sequence.length) && (this.curTree.leaves.length == 0);
}

/**
 * Check if we are at the beginning of the game tree.
 *
 * @returns {boolean} True if at beginning, false if not.
 */
JGONavigator.prototype.isAtBeginning = function() {
	return (this.seqPos == 0) && (this.treeStack.length == 0);
}

function jgo_explodeSGFList(prop) {
	var coords = [];
	
	$.each(prop, function(i, val) {
		if(val.indexOf(":") == -1) { // single coordinate
			coords.push(new JGOCoordinate(val));
		} else {
			var tuple = v.split(":"), c1, c2;
			
			c1 = new JGOCoordinate(tuple[0]);
			c2 = new JGOCoordinate(tuple[1]);
			coord = new JGOCoordinate();
			
			for(coord.i=c1.i; coord.i<=c2.i; ++coord.i)
				for(coord.j=c1.j; coord.j<=c2.j; ++coord.j)
					coords.push(coord.copy());
		}
	});
	
	return coords;
}

/**
 * Execute an SGF node on a board.
 *
 * @param {object} node The SGF node, as associative array object, as created by jgo_parseSGF
 * @param {JGOBoard} board The board to apply node changes to
 * @param {boolean} moveMarker Set to true if you want a last move marker on move nodes.
 * @returns {object} Information on the move made: {type: <t>, captures: <n>, diff: <diff>, markers: {...} }
 * where t=0 for info node, t=1 for a move, and t=2 for a setup, and diff contains changes made (see JGOBoard.getDiff). 
 * "markers" property is a coordinate-marker -map that can be fed to visible board's mark() function
 */
function jgo_executeSGFNode(node, board, moveMarker) {
	var coord, own, enemy, ret = {type: 0, captures: [0,0], diff: [[],[],[]], markers: {}};

	// check for move type nodes
	if("B" in node && node["B"].length == 1) {
		if(node["B"][0].length == 2)
			coord = new JGOCoordinate(node["B"][0]);
		own = JGO_BLACK;
		enemy = JGO_WHITE;
		ret.type = 1;
		
		if(moveMarker)
			ret.markers['0'] = [coord.copy()]; // make a copy just in case
	} else if("W" in node && node["W"].length == 1) {
		if(node["W"][0].length == 2)
			coord = new JGOCoordinate(node["W"][0]);
		own = JGO_WHITE;
		enemy = JGO_BLACK;
		ret.type = 1;
		
		if(moveMarker)
			ret.markers['0'] = [coord.copy()]; // make a copy just in case
	}
	
	var markerNodes = {"TW": ',', "TB": '.', "CR": '0', "TR": '/', "MA": '*', "SQ": '#'};
	
	$.each(markerNodes, function(markerNode, marker) {
		if(markerNode in node)
			ret.markers[marker] = jgo_explodeSGFList(node[markerNode]);
	});
	
	if("LB" in node) {
		$.each(node["LB"], function(i, v) {
			var tuple = v.split(":");
			
			if(!(tuple[1] in ret.markers))
				ret.markers[tuple[1]] = [];
				
			ret.markers[tuple[1]].push(new JGOCoordinate(tuple[0]));
		});
	}
	
	// check for setyp type nodes, and assume not invalid SGF mixed with move node
	if(ret.type != 1) {
		var setupNodes = {"AB": JGO_BLACK, "AW": JGO_WHITE, "AE": JGO_CLEAR};
		
		$.each(setupNodes, function(setupNode, color) {
			if(setupNode in node) {
				var coords = jgo_explodeSGFList(node[setupNode]), i;
				
				for(i=0; i<coords.length; i++) {
					coord = coords[i];
					board.set(coord, color);
					ret.diff[color].push(coord);
				}
				
				ret.type = 2;
			}
		});
	}
	
	if(ret.type != 1 || coord == undefined) // for general and setup nodes and "pass"
		return ret;
		
	board.set(coord, own); // put own stone on board
	ret.diff[own].push(coord);
		
	var enemies = board.getAdjacent(coord, enemy);
	
	$.each(enemies, function(i, eCoord) {
		if(board.get(eCoord) != JGO_CLEAR) { // check that we have not already removed these enemy stones
			var group = board.getGroup(eCoord);
			
			if(!board.hasLiberties(group)) {
				board.set(group, JGO_CLEAR);
				ret.diff[JGO_CLEAR] = ret.diff[JGO_CLEAR].concat(group);
				ret.captures[(own == JGO_BLACK ? 0 : 1)] += group.length; // black's captures go to prisoners[0], white's captures to prisoners[1]
			}
		}
	});
	
	// suicide not allowed in these rules, so no need to check that...
	return ret;
}

/**
 * Get game information from a node. 
 * Only a subset of the full game info is implemented - see below for mapped properties.
 *
 * @param {object} The SGF node game tree
 * @returns {object} Information on the game with keys that were found from root node
 */
function jgo_getSGFGameInfo(node) {
	var info = {};
	var fields = {"AN": "annotator", "CP": "copyright", "DT": "date", "EV": "event", "GN": "gameName", "OT": "overtime", 
	"RO": "round", "RE": "result", "RU": "rules", "SO": "source", "TM": "time", "PC": "location",
	"PB": "black", "PW": "white", "BR": "blackRank", "WR": "whiteRank", "BT": "blackTeam", "WT": "whiteTeam"};
	
	$.each(fields, function(sgfKey, infoKey) {
		if(sgfKey in node)
			info[infoKey] = node[sgfKey];
	});
	
	return info;
}

/**
 * Parse SGF string into object tree representation: 
 *
 * tree = { sequence: [ <node(s)> ], leaves: [ <subtree(s), if any> ] }
 *
 * Each node is an object containing property identifiers and associated values:
 *
 * node = {"B": ["nn"], "C": ["This is a comment"]}
 *
 * @param {String} sgf The SGF in string format, whitespace allowed.
 */
function jgo_parseSGF(sgf) {
	var tokens,	i, len, token, // for loop vars 
		lastBackslash = false, // flag to note if last string ended in escape
		bracketOpen = -1, // the index where bracket opened
		processed = [];
	
	if("a~b".split(/(~)/).length === 3) {
		tokens = sgf.split(/([\[\]\(\);])/); // split into an array at "[", "]", "(", ")", and ";", and include separators in array	
	} else { // Thank you IE for not working
		var blockStart = 0, delimiters = "[]();";
		
		tokens = [];
		
		for(i=0, len=sgf.length; i<len; ++i) {
			if(delimiters.indexOf(sgf.charAt(i)) !== -1) {
				if(blockStart < i)
					tokens.push(sgf.substring(blockStart, i));
				tokens.push(sgf.charAt(i));
				blockStart = i+1;
			}
		}
		
		if(blockStart < i) // leftovers
			tokens.push(sgf.substr(blockStart, i));
	}

	// process through tokens and push everything into processed, but merge stuff between square brackets into one element, unescaping escaped brackets
	// i.e. ["(", ";", "C", "[", "this is a comment containing brackets ", "[", "\\", "]", "]"] becomes:
	// ["(", ";", "C", "[", "this is a comment containing brackets []]"]
	// after this transformation, it's just "(", ")", ";", "ID", and "[bracket stuff]" elements in the processed array
	for(i=0, len=tokens.length; i<len; ++i) {
		token = tokens[i];

		if(bracketOpen == -1) { // handling elements outside property values (i.e. square brackets)
			token = jQuery.trim(token); // trim whitespace, it is irrelevant here
			if(token == "[") // found one
				bracketOpen = i;
			else if(token != "") // we are outside brackets, so just push everything nonempty as it is into 'processed'
				processed.push(token);
		} else { // bracket is open, we're now looking for a ] without preceding \
			if(token != "]") { // a text segment
				lastBackslash = (token.charAt(token.length-1) == "\\"); // true if string ends in \
			} else { // a closing bracket
				if(lastBackslash) { // it's escaped - we continue
					lastBackslash = false; 
				} else { // it's not escaped - we close the segment
					processed.push(tokens.slice(bracketOpen, i+1).join('').replace(/\\\]/g, "]")); // push the whole thing including brackets, and unescape the inside closing brackets
					bracketOpen = -1;
				}
			}	
		}
	}
		
	// basic error checking
	if(processed.length == 0) {
		jgo_errorStr = "SGF was empty!";
		return false;
	} else if(processed[0] != "(" || processed[1] != ";" || processed[processed.length-1] != ")") {
		jgo_errorStr = "SGF did not start with \"(;\" or end with \")\"!";
		return false;
	}
		
	// collect "XY", "[AB]", "[CD]" sequences (properties) in a node into {"XY": ["AB", "CD"]} type of associative array
	// effectively parsing "(;GM[1]FF[4];B[pd])" into ["(", {"GM": ["1"], "FF": ["4"]}, {"B": ["pd"]}, ")"]
	
	// start again with "tokens" and process into "processed"
	tokens = processed; 
	processed = [];
	
	var node, propertyId = ""; // node under construction, and current property identifier
	
	// the following code is not strict on format, so let's hope it's well formed
	for(i=0, len=tokens.length; i<len; ++i) {
		token = tokens[i];
		
		if(token == "(" || token == ")") { 
			if(node) { // flush and reset node if necessary
				processed.push(node);
				node = undefined;
			}
			
			processed.push(token); // push this token also
		} else if(token == ";") { // new node
			if(node) // flush if necessary
				processed.push(node);
			
			node = {}; propertyId = ""; // initialize the new node
		} else { // it's either a property identifier or a property value
			if(token.indexOf("[") != 0) { // it's property identifier
				if(propertyId != "" && node[propertyId].length == 0) { // last property was missing value
					jgo_errorStr = "Missing property value at " + token + "!";
					return false;
				}
				
				if(token in node) { // duplicate key
					jgo_errorStr = "Duplicate property identifier " + token + "!";
					return false;
				}
				
				propertyId = token;
				node[propertyId] = []; // initialize new property with empty value array
			} else { // it's property value
				if(propertyId == "") { // we're missing the identifier
					jgo_errorStr = "Missing property identifier at " + token + "!";
					return false;
				}
				
				node[propertyId].push(token.substring(1, token.length-1)); // remove enclosing brackets
			}
		}
	}

	tokens = processed; 

	// finally, construct a game tree from "(", ")", and sequence arrays - each leaf is {sequence: [ <list of nodes> ], leaves: [ <list of leaves> ]}
	var stack = [], currentRoot = {sequence: [], leaves: []}, lastRoot; // we know first element already: "("
	
	for(i=1, len=tokens.length; i<len-1; ++i) {
		token = tokens[i];
		
		if(token == "(") { // enter a subleaf
			if(currentRoot.sequence.length == 0) { // consecutive parenthesis without node sequence in between will throw an error
				jgo_errorStr = "SGF contains a game tree without a sequence!";
				return false;
			}
			stack.push(currentRoot); // save current leaf for when we return
			currentRoot = {sequence: [], leaves: []};
		} else if(token == ")") { // return from subleaf
			if(currentRoot.sequence.length == 0) { // consecutive parenthesis without node sequence in between will throw an error
				jgo_errorStr = "SGF contains a game tree without a sequence!";
				return false;
			}
			lastRoot = currentRoot;
			currentRoot = stack.pop();
			currentRoot.leaves.push(lastRoot);
		} else { // if every "(" is not followed by exactly one array of nodes (as it should), this code fails
			currentRoot.sequence.push(token);
		}
	}
	
	if(stack.length > 0) {
		jgo_errorStr = "Invalid number of parentheses in the SGF!";
		return null;
	}
	
	return currentRoot;
}
